%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
% This function extracts five sets of features:
% 1. Intensity-based features
% 2. Textural features based on Gabor filters
% 3. Textural features based on local binary patterns
% 4. Textural features based on co-occurrence matrices
% 5. Graph-based features
%
% It extracts the intensity-based and textural features on one-channel
% image given as a parameter (it can be a gray-level image or the
% hematoxylin channel)
%
% It constructs a graph, considering the cell centroids as nodes and
% assigning edges between these nodes using Delaunay triangulation
%
%
% oneChannelImage   : Gray-scale or normalized gray-scale or hematoxylin
%                     channel. But no RGB image.
% centx             : X coordinates of the centroids (rows in an image)
% centy             : Y coordinates of the centroids (columns in an image)
% windowHalfSize    : Half size of the window to be cropped around each
%                     centroid. Total window has the size of
%                     2 * windowHalfSize + 1
% minIntensity      : For contrast streching, use 0 to ignore
% maxIntensity      : For contrast streching, use 255 to ignore
% binNo             : Intensity-based and cooccurrence-based features are
%                     calculated on the quantized image. This is the number
%                     of bins to be used.
% coocDist          : The cooccurrence matrices are calculated based on
%                     this distance. They are calculated this distance and
%                     eight different orientations and then summed up
%                     before calculating the features.
% wthr              : Delaunay triangulation may have too long edges for
%                     the nodes close to the image boundaries. To get rid
%                     of such edges, use positive wthr. It is a threshold
%                     to understand which nodes can be considered as close
%                     to boundary. Use -1 to ignore this property.
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function [texturalFeatures, graphFeatures] = extractFeatures(oneChannelImage, centx, centy, windowHalfSize, ...
                                                             minIntensity, maxIntensity, binNo, coocDist, wthr, calculateGraph)

% It calculates those maps for the entire image and then accumulates the
% features of the pixels falling in the window cropped around each cell
% Gabor filter parameters are those suggested by the original implementer
quanImage = quantizeArray(oneChannelImage, minIntensity, maxIntensity, binNo);
LBPimage = calculateLBP(oneChannelImage);
convMagnitude = convolveWithGabor(oneChannelImage, 4, 6, 3, 2, 0.65); 
texturalFeatures = extractTexturalFeatures(centx, centy, windowHalfSize, oneChannelImage, quanImage, LBPimage, convMagnitude, coocDist);

if calculateGraph == 1
    adjM = constructTriangulationGraph(centx, centy, size(oneChannelImage, 1), size(oneChannelImage, 2), wthr);
    graphFeatures = extractGraphFeatures(adjM, centx, centy);
else
    graphFeatures = 0;
end
%plotGraph(img, centx, centy, adjM);

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function [binStart, binEnd] = findBinEndPoints(minIntensity, maxIntensity, binNo)

binStart = zeros(binNo, 1);
binEnd = zeros(binNo, 1);
binSize = round((maxIntensity - minIntensity) / binNo);

binStart(1) = minIntensity;
binEnd(binNo) = maxIntensity;
for i = 1 : binNo - 1
    binEnd(i) = binStart(i) + binSize - 1;
    binStart(i + 1) = binEnd(i) + 1;
end

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function arr = quantizeArray(img, minIntensity, maxIntensity, binNo)

[binStart, ~] = findBinEndPoints(minIntensity, maxIntensity, binNo);
[dx, dy] = size(img);
arr = zeros(dx, dy);
for i = 1 : dx
    for j = 1 : dy
        selBin = 1;
        for k = 2 : binNo
            if img(i, j) >= binStart(k)
                selBin = k;
            else
                break
            end
        end
        arr(i, j) = selBin;
    end
end

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function convMagnitude = convolveWithGabor(img, nscale, norient, minWaveLength, mult, sigmaOnf)

[EO, BP] = gaborconvolve(img, nscale, norient, minWaveLength, mult, sigmaOnf);
for i = 1 : nscale
    convMagnitude{i} = abs(EO{i, 1});
    for j = 2 : norient
        convMagnitude{i} = convMagnitude{i} + abs(EO{i, j});
    end
    convMagnitude{i} = convMagnitude{i} / norient;
end
% abs(EO): magnitude of the convolution
% angle(EO): phase angles
% BP: bandpass images

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function LBPimage = calculateLBP(img)

[dx, dy] = size(img);
LBPimage = zeros(dx, dy);
for i = 2 : dx-1
    for j = 2 : dy-1
        str = zeros(8, 1);
        if img(i, j) > img(i, j+1),   str(1) = 1; end
        if img(i, j) > img(i+1, j+1), str(2) = 1; end
        if img(i, j) > img(i+1, j),   str(3) = 1; end
        if img(i, j) > img(i+1, j-1), str(4) = 1; end
        if img(i, j) > img(i, j-1),   str(5) = 1; end
        if img(i, j) > img(i-1, j-1), str(6) = 1; end
        if img(i, j) > img(i-1, j),   str(7) = 1; end
        if img(i, j) > img(i-1, j+1), str(8) = 1; end
        
        for k = 1 : 8
            LBPimage(i, j) = LBPimage(i, j) + str(k) * 2^(8-k);
        end
    end
end

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%



