function allFeatures = extractTexturalFeatures(centx, centy, hsize, img, quanImg, LBPimg, convMagnitude, coocDist)

centNo = length(centx);
binNo = max(max(quanImg));
dx = size(img, 1);
dy = size(img, 2);
img = double(img);
for i = 1 : centNo
    [minx, maxx, miny, maxy] = findBoundingBox(centx(i), centy(i), hsize, dx, dy);
    
    selImg  = img(minx:maxx, miny:maxy);
    selQuan = quanImg(minx:maxx, miny:maxy);
    selLBP  = LBPimg(minx:maxx, miny:maxy);

    features = extractIntensityFeatures(selImg, selQuan, binNo);
    features = extractCooccurrenceFeatures(features, selQuan, binNo, coocDist);
    features = extractLBPFeatures(features, selLBP);

    roundPrecision = 0;
    features = extractGaborFeatures(features, convMagnitude, minx, maxx, miny, maxy, roundPrecision);
    
    allFeatures(i, 1) = features;
end

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function features = extractIntensityFeatures(img, quanImg, binNo)

pixNo = size(img, 1)* size(img, 2);
pixArray = reshape(img, [pixNo 1]);
features.MinIntensity = min(pixArray);
features.MaxIntensity = max(pixArray);
features.AvgIntensity = mean(pixArray);
features.StdIntensity = std(pixArray);
features.SkewnessIntensity = skewness(pixArray);
features.KurtosisIntensity = kurtosis(pixArray);

quanArray = reshape(quanImg, [pixNo 1]);
histogram = calculateHistogram(quanArray, binNo);
histogram = histogram / pixNo;
features.Entropy = calculateEntropy(histogram);

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function features = extractCooccurrenceFeatures(features, quanImg, binNo, coocDist)

accM = calculateAccumulatedCooccurrenceMatrix(quanImg, binNo, coocDist);

if sum(sum(accM))
    accM = accM / sum(sum(accM));
    coocFeat = calculateCooccurrenceFeatures(accM);

    features.CoocMaximumProbability = coocFeat.MaximumProbability;
    features.CoocEntropy = coocFeat.Entropy;
    features.CoocAngularSecondMoment = coocFeat.AngularSecondMoment;
    features.CoocInverseDifferenceMoment = coocFeat.InverseDifferenceMoment;
    features.CoocContrast = coocFeat.Contrast;
    features.CoocCorrelation = coocFeat.Correlation;
else
    features.CoocMaximumProbability = 0;
    features.CoocEntropy = 0;
    features.CoocAngularSecondMoment = 0;
    features.CoocInverseDifferenceMoment = 0;
    features.CoocContrast = 0;
    features.CoocCorrelation = 0;
end

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function features = extractGaborFeatures(features, convMagnitude, minx, maxx, miny, maxy, roundPrecision)

nscale = size(convMagnitude, 2);
for i = 1 : nscale
    currImg = convMagnitude{i};
    selImg = currImg(minx:maxx, miny:maxy);
    selArray = reshape(selImg, [size(selImg, 1)* size(selImg, 2)  1]);
    features.GaborAvg(i) = mean(selArray);
    features.GaborStd(i) = std(selArray);
    features.GaborMin2Max(i) = min(selArray) / max(selArray);
    features.GaborMode(i) = mode(round(selArray, roundPrecision));
end

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function features = extractLBPFeatures(features, LBPimage)

pixNo = size(LBPimage, 1)* size(LBPimage, 2);
pixArray = reshape(LBPimage, [pixNo 1]);

features.LBPAvg = mean(pixArray);
features.LBPStd = std(pixArray);
features.LBPMode = mode(pixArray);

histogram = calculateHistogram(pixArray+1, 256);
histogram = histogram / pixNo;
features.LBPEntropy = calculateEntropy(histogram);

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function accM = calculateAccumulatedCooccurrenceMatrix(quanImg, binNo, coocDist)

M1 = calculateCooccurrenceMatrix(quanImg, binNo, coocDist, 0);
M3 = calculateCooccurrenceMatrix(quanImg, binNo, 0, coocDist);
M5 = calculateCooccurrenceMatrix(quanImg, binNo, -coocDist, 0);
M7 = calculateCooccurrenceMatrix(quanImg, binNo, 0, -coocDist);

M2 = calculateCooccurrenceMatrix(quanImg, binNo, coocDist, coocDist);
M8 = calculateCooccurrenceMatrix(quanImg, binNo, coocDist, -coocDist);
M4 = calculateCooccurrenceMatrix(quanImg, binNo, -coocDist, coocDist);
M6 = calculateCooccurrenceMatrix(quanImg, binNo, -coocDist, -coocDist);

accM = M1 + M2 + M3 + M4 + M5 + M6 + M7 + M8;

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function M = calculateCooccurrenceMatrix(quanImg, binNo, di, dj)

[dx, dy] = size(quanImg);

M = zeros(binNo, binNo);
for i = 1 : dx
    for j = 1 : dy
        if i + di >= 1 && i + di <= dx && j + dj >= 1 && j + dj <= dy
            M(quanImg(i, j), quanImg(i+di, j+dj)) = M(quanImg(i, j), quanImg(i+di, j+dj)) + 1;
        end
    end
end

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function feat = calculateCooccurrenceFeatures(accM)

[dimx, dimy] = size(accM);
feat.MaximumProbability = max(max(accM));
feat.Entropy = calculateEntropy(accM);

% AngularSecondMoment
feat.AngularSecondMoment = 0;
for i = 1 : dimx
    for j = 1 : dimy
        feat.AngularSecondMoment = feat.AngularSecondMoment + accM(i, j) * accM(i, j);
    end
end
% InverseDifferenceMoment
feat.InverseDifferenceMoment = 0;
for i = 1 : dimx
    for j = 1 : dimy
        feat.InverseDifferenceMoment = feat.InverseDifferenceMoment + accM(i, j) / (1 + (i-j)*(i-j)); 
    end
end
% Contrast
feat.Contrast = 0;
for i = 1 : dimx
    for j = 1 : dimy
        feat.Contrast = feat.Contrast + accM(i, j) * (i-j)*(i-j); 
    end
end
% Correlation
Ni = zeros(dimx, 1);
for i = 1 : dimx
    for j = 1 : dimy
        Ni(i, 1) = Ni(i, 1) + accM(i, j);
    end
end
Nj = zeros(dimy, 1);
for i = 1 : dimy
    for j = 1 : dimx
        Nj(i, 1) = Nj(i, 1) + accM(j, i);
    end
end
mu_i = mean(Ni);
s_i  = std(Ni);
mu_j = mean(Nj);
s_j  = std(Nj);
feat.Correlation = 0;
if s_i * s_j
    for i = 1 : dimx
        for j = 1 : dimy
            feat.Correlation = feat.Correlation + (i * j * accM(i, j) - mu_i * mu_j) / (s_i * s_j);
        end
    end
end

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function [minx, maxx, miny, maxy] = findBoundingBox(onex, oney, hsize, dx, dy)

minx = onex - hsize;
maxx = onex + hsize;
miny = oney - hsize;
maxy = oney + hsize;

if minx < 1,  minx = 1;  end
if maxx > dx, maxx = dx; end
if miny < 1,  miny = 1;  end
if maxy > dy, maxy = dy; end

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function entropy = calculateEntropy(arr)

[ax, ay] = size(arr);
entropy = 0;
for i = 1 : ax
    for j = 1 : ay
        if arr(i, j) > 0
            entropy = entropy - arr(i, j) * log(arr(i, j));
        end
    end
end

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
function histogram = calculateHistogram(arr, binNo)

len = length(arr);
histogram = zeros(binNo, 1);
for i = 1 : len
    histogram(arr(i)) = histogram(arr(i)) + 1;
end

end
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%
%--------------------------------------------------------------------%

